#ifndef CYGONCE_HAL_HAL_ARCH_H
#define CYGONCE_HAL_HAL_ARCH_H

//==========================================================================
//
//      hal_arch.h
//
//      Architecture specific abstractions
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Hao Tseng
// Contributors: None
// Date:         2003-11-12
// Purpose:      Define architecture abstractions
// Usage:        #include <cyg/hal/hal_arch.h>
// 
//####DESCRIPTIONEND####
//
//==========================================================================
#include <cyg/infra/cyg_type.h>
#include <pkgconf/hal.h>

#include <pkgconf/system.h>
#include CYGBLD_HAL_PLATFORM_H
#include <cyg/hal/arch.inc>
#include <cyg/hal/hal_intr.h>


//--------------------------------------------------------------------------
// (1) Struct HAL_SavedRegisters
	typedef struct 
	{
		cyg_uint32	Status;		// saved status register for saving CWP
		cyg_uint32	IStatus;	// saved IStatus register for interrupt, each context switching happens in interrupt(vsr).        
		cyg_uint32  g[8] ;                                  /* Golbal regs */
		cyg_uint32  o[8] ;                                  /* Output regs */
		cyg_uint32  l[8] ;                                  /* Local regs */        
		cyg_uint32  i[8] ;                                  /* Input regs */
	} HAL_SavedRegisters;


	// memory layout sequence of HAL_SavedRegisters    
	// STATUS --> 0
	// ISTATUS --> 1
	// %g0 --> 2
	// %g1 --> 3
	// %g2 --> 4
	// %g3 --> 5
	// %g4 --> 6
	// %g5 --> 7
	// %g6 --> 8
	// %g7 --> 9
	// %o0 --> 10
	// %o1 --> 11
	// %o2 --> 12
	// %o3 --> 13
	// %o4 --> 14
	// %o5 --> 15
	// %o6 --> 16
	// %o7 --> 17
    // %l0 --> 18
	// %l1 --> 19
	// %l2 --> 20
	// %l3 --> 21
	// %l4 --> 22
	// %l5 --> 23
	// %l6 --> 24
	// %l7 --> 25
	// %i0 --> 26
	// %i1 --> 27
	// %i2 --> 28
	// %i3 --> 29
	// %i4 --> 30
	// %i5 --> 31
	// %i6 --> 32
	// %i7 --> 33

//--------------------------------------------------------------------------
// (2) HAL_LSBIT_INDEX(), HAL_MSBIT_INDEX()

	// Bit manipulation macro
	// both implemted in platform HAL
	// get the "1" of the least significant position of  mask
	externC cyg_uint32 hal_lsbit_index(cyg_uint32 mask);

	// get the "1" of the most significant position of  mask
	externC cyg_uint32 hal_msbit_index(cyg_uint32 mask);

	#define HAL_LSBIT_INDEX(index, mask) index = hal_lsbit_index(mask);
	#define HAL_MSBIT_INDEX(index, mask) index = hal_msbit_index(mask);

//--------------------------------------------------------------------------
// (3) HAL_THREAD_INIT_CONTEXT() 
//     HAL_THREAD_LOAD_CONTEXT() 
//     HAL_THREAD_SWITCH_CONTEXT()

// Context Initialization
// Initialize the context of a thread.
// Arguments:
// _sparg_ name of variable containing current sp, will be written with new sp
// _thread_ thread object address, passed as argument to entry point
// _entry_ entry point address.
// _id_ bit pattern used in initializing registers, for debugging.
#define HAL_NIOS_GET_INIT_STATUS_REG_VALUE(_reg0_)    \
{                                               \
    asm volatile (                              \
        "pfx 2 \n                               \
        rdctl  %0 \n                            \
        pfx    %%hi(0x3e0) \n                   \
        and    %0, %%lo(0x3e0) \n               \
        lsri   %0,1 \n                          \
        pfx    %%hi(0x7e00) \n                  \
        or     %0, %%lo(0x7e00)"                \
        : "=r"(_reg0_)                          \
        );                                      \
}

#define HAL_THREAD_INIT_CONTEXT( _sparg_, _thread_, _entry_, _id_ )                     \
{                                                                                       \
    register cyg_uint32 _sp_ = ((cyg_uint32)_sparg_)-4;/* Reserve 4 bytes to stroe id */\
    register HAL_SavedRegisters *_regs_;                                                \
    cyg_uint32 _i_;                                                                     \
    _regs_ = (HAL_SavedRegisters *)((_sp_) - sizeof(HAL_SavedRegisters));               \
    *((cyg_uint32 *)_sp_) = (cyg_uint32)_id_;                                           \
    (_regs_)->o[7] = (cyg_uint32)(_entry_);          /* o7 = thread's entry point */    \
    (_regs_)->o[6] = (cyg_uint32)(_sp_);             /* o6 = thread's sp value */       \
    (_regs_)->o[0] = (cyg_uint32)(_thread_);         /* o0 = thread's parameters */     \
    HAL_NIOS_GET_INIT_STATUS_REG_VALUE(_i_);                                            \
    _i_ |= (cyg_uint32)(0x10000);        /* set ICache */                               \
    _i_ |= (cyg_uint32)(0x20000);        /* set DCache */                               \
    (_regs_)->Status = (_regs_)->IStatus = _i_;                                         \
    _sparg_ = (CYG_ADDRWORD)_sp_;                                                       \
}

// Context switch macros.
// The arguments are pointers to locations where the stack pointer
// of the current thread is to be stored, and from where the sp of the
// next thread is to be fetched.
externC void hal_thread_switch_context( CYG_ADDRESS to, CYG_ADDRESS from );
externC void hal_thread_load_context( CYG_ADDRESS to )
	__attribute__ ((noreturn));

#define HAL_THREAD_SWITCH_CONTEXT(_fspptr_,_tspptr_)                    \
		hal_thread_switch_context((CYG_ADDRESS)_tspptr_, (CYG_ADDRESS)_fspptr_);
                                  
#define HAL_THREAD_LOAD_CONTEXT(_tspptr_)                               \
		hal_thread_load_context( (CYG_ADDRESS)_tspptr_ );

//--------------------------------------------------------------------------
// (3) HAL_REORDER_BARRIER()

// Execution reorder barrier.
// When optimizing the compiler can reorder code. In multithreaded systems
// where the order of actions is vital, this can sometimes cause problems.
// This macro may be inserted into places where reordering should not happen.
// The "memory" keyword is potentially unnecessary, but it is harmless to
// keep it.

#define HAL_REORDER_BARRIER() asm volatile ( "" : : : "memory" )

//--------------------------------------------------------------------------
// (4) HAL_BREAKPOINT(), HAL_BREAKINST, HAL_BREAKINST_SIZE

// Breakpoint support
// HAL_BREAKPOINT() is a code sequence that will cause a breakpoint to
// happen if executed.
// HAL_BREAKINST is the value of the breakpoint instruction and
// HAL_BREAKINST_SIZE is its size in bytes.
// HAL_BREAKINST_TYPE is the type.

#define HAL_BREAKPOINT(_label_)                 \
asm volatile (" .globl  " #_label_ ";"          \
              #_label_":"                       \
              " trap   3"                       \
    );

#define HAL_BREAKINST           0x7903

#define HAL_BREAKINST_SIZE      2

#define HAL_BREAKINST_TYPE      cyg_uint16

#define NOP_INSTR               0x3000

//--------------------------------------------------------------------------
// (5) HAL_GET_GDB_REGISTERS()
//      HAL_SET_GDB_REGISTERS()
//      HAL_THREAD_GET_SAVED_REGISTERS()

// Default to a 32 bit register size for GDB register dumps.
#ifndef CYG_HAL_GDB_REG
#define CYG_HAL_GDB_REG CYG_WORD32
#endif

// Translate a stack pointer as saved by the thread context macros above into
// a pointer to a HAL_SavedRegisters structure.
#define HAL_THREAD_GET_SAVED_REGISTERS( _sp_, _regs_ )          \
        (_regs_) = (HAL_SavedRegisters *)(_sp_)
        
#define HAL_GET_GDB_REGISTERS( _aregval_, _regs_ )              \
{                                                               \
    CYG_HAL_GDB_REG *_regval_ = (CYG_HAL_GDB_REG *)(_aregval_); \
    CYG_HAL_GDB_REG *_oregs_ = (CYG_HAL_GDB_REG *)(_regs_);     \
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 34; _i_++ )                             \
        _regval_[_i_] = _oregs_[_i_];                           \
}

#define HAL_SET_GDB_REGISTERS( _regs_ , _aregval_ )             \
{                                                               \
    CYG_HAL_GDB_REG *_regval_ = (CYG_HAL_GDB_REG *)(_aregval_); \
    CYG_HAL_GDB_REG *_oregs_ = (CYG_HAL_GDB_REG *)(_regs_);     \
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 34; _i_++ )                             \
        _oregs_[_i_] = _regval_[_i_];                           \
}


//--------------------------------------------------------------------------
// (6) hal_jmp_buf, hal_setjmp(), hal_longjmp()

// HAL setjmp
#define CYGARC_JMP_BUF_SIZE SAVED_REGS_SIZE

typedef cyg_uint32 hal_jmp_buf[ CYGARC_JMP_BUF_SIZE / 4 ];

externC int hal_setjmp(hal_jmp_buf env);
externC void hal_longjmp(hal_jmp_buf env, int val);

//--------------------------------------------------------------------------
// (7) HAL_IDLE_THREAD_ACTION()

externC void hal_idle_thread_action(cyg_uint32 loop_count);
#define HAL_IDLE_THREAD_ACTION(_count_) hal_idle_thread_action(_count_)

//--------------------------------------------------------------------------
// (8) CYGNUM_HAL_STACK_SIZE_MINIMUM, CYGNUM_HAL_STACK_SIZE_TYPICAL

#define CYGNUM_HAL_STACK_FRAME_SIZE (4 * 35)	// 35 words --- 140 bytes
#define CYGNUM_HAL_STACK_CONTEXT_SIZE (4 * 35)	// 35 words --- 140 bytes
#define CYGNUM_HAL_STACK_INTERRUPT_SIZE	((208 * 4) + 2 * CYGNUM_HAL_STACK_FRAME_SIZE)	// 1112 bytes

#define CYGNUM_HAL_STACK_SIZE_MINIMUM	(4 * CYGNUM_HAL_STACK_INTERRUPT_SIZE + 2 * CYGNUM_HAL_STACK_FRAME_SIZE)	// 4728 bytes
#define CYGNUM_HAL_STACK_SIZE_TYPICAL	(CYGNUM_HAL_STACK_SIZE_MINIMUM + 8 * CYGNUM_HAL_STACK_FRAME_SIZE)	// 5848 bytes

//--------------------------------------------------------------------------
// *(9) Memory access macros


//--------------------------------------------------------------------------
// (10) Global pointer save/restore macros
//
// Macros for switching context between two eCos instances (jump from
// code in ROM to code in RAM or vice versa).
#define CYGARC_HAL_SAVE_GP()
#define CYGARC_HAL_RESTORE_GP()


////--------------------------------------------------------------------------
//// (11) Virtual Vector Table
//#ifdef CYGSEM_HAL_VIRTUAL_VECTOR_SUPPORT
// #include <cyg/hal/hal_if.h>
// #define hal_virtual_vector_table ((CYG_ADDRWORD *)(CYGHWR_NIOS_VECTOR_TABLE_ADDRESS - (CYGNUM_CALL_IF_TABLE_SIZE * 4)))
//#endif

#endif // CYGONCE_HAL_HAL_ARCH_H
// End of hal_arch.h
