//=============================================================================
//	context.S
//
//	NIOS context switch code
//
//=============================================================================
//###ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//###ECOSGPLCOPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s): 	Hao Tseng
// Contributors:
// Date:	
// Purpose:	NIOS context switch code
// Description:	This file contains implementations of the thread context 
//		        switch routines. It also contains the longjmp() and setjmp()
//              routines.
//
//####DESCRIPTIONEND####
//
//=============================================================================

#include <pkgconf/hal.h>
#include <cyg/hal/arch.inc>
#include <cyg/hal/nios_imacros.inc>

#ifdef at
#undef at
#endif

// hal_thread_switch_context
// switch thread contexts
// %o0 = address of sp of next thread to excute
// %o1 = address of sp save location of current thread

	.global hal_thread_switch_context
	// .global makes the symbol visible to ld.
hal_thread_switch_context:
	
	// Disable Interrupt
	PFX 8
	WRCTL %g0
	NOP

	// the parameter passed back (location store the stack pointer)	
	// Note: %o1 passed from caller is the address of where the old thread pointer to be stored
    ST  [%o1],%sp
    
	// save the register set of current thread, we are in the same register window as the caller
    SUBIP %sp, SAVED_REGS_SIZE

	// save the %g0~7 registers of the old thread
	STS [%sp, 2], %g0
	STS [%sp, 3], %g1
	STS [%sp, 4], %g2
	STS [%sp, 5], %g3
	STS [%sp, 6], %g4
	STS [%sp, 7], %g5
	STS [%sp, 8], %g6
	STS [%sp, 9], %g7
    
	// save the %o0~7 registers of the old thread
	STS [%sp, 10], %o0
	STS [%sp, 11], %o1
	STS [%sp, 12], %o2
	STS [%sp, 13], %o3
	STS [%sp, 14], %o4
	STS [%sp, 15], %o5
	STS [%sp, 17], %o7
    
    MOV %g0,%sp                     ;save original sp to HAL_SavedRegisters
    ADDIP %g0, SAVED_REGS_SIZE
	STS [%sp, 16], %g0

	// save the %l0~7 registers of the old thread
	STS [%sp, 18], %l0    
	STS [%sp, 19], %l1
	STS [%sp, 20], %l2
	STS [%sp, 21], %l3
	STS [%sp, 22], %l4
	STS [%sp, 23], %l5
	STS [%sp, 24], %l6
	STS [%sp, 25], %l7

	// save the %i0~7 registers of the old thread
	STS [%sp, 26], %i0
	STS [%sp, 27], %i1
	STS [%sp, 28], %i2
	STS [%sp, 29], %i3
	STS [%sp, 30], %i4
	STS [%sp, 31], %i5
	STS [%sp, 32], %i6
	STS [%sp, 33], %i7
    
	// save "Status" of the old thread
	RDCTL %g0			
	STS [%sp, 0], %g0

	// save "IStatus" of the old thread
	PFX 1
	RDCTL %g0
	STS [%sp, 1], %g0
	
	// save the passed parameter %o0(the new thread stack pointer) into %g0
	MOV %g0, %o0
    
    // backup_all_cwps_to_stack(), %sp = stack
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    ADDIP   %sp, SAVED_REGS_SIZE
    .extern backup_all_cwps_to_stack
    _BSR    backup_all_cwps_to_stack, %g7
    NOP
    
	// reconstruct the passed parameter in %o0
	MOV %o0, %g0

// all "*old*" thread's register windows have been saved 
// now we can restore the "*new*" thread's register window
// by droping through into "hal_thread_load_context"

// ------------------------------------------------------------------------------
// hal_thread_load_context
// Load thread context
// %o0 = address of sp of next thread to execute
// Note that this function is also the second half of hal_thread_switch_context
// and is simply dropped into from it.
	.global hal_thread_load_context
hal_thread_load_context:

	// Disable Interrupt
	PFX 8
	WRCTL %g0
	NOP
    
    // Get the sp that have next thread's context
    LD      %g6, [%o0]               ; let g6 = stack pointer
    SUBIP   %g6, SAVED_REGS_SIZE     ; %g6 = NSP

    // Restore STATUS first, and let %g0 = STATUS
    LD     %g0, [%g6]           ; get STATUS into %g0
    WRCTL   %g0                 ; write STATUS value
    NOP

    PFX     32
    LD      %fp,[%g6]
    MOV     %sp, %g6
    ADDIP   %sp, SAVED_REGS_SIZE    
    
    // restore_all_cwps_from_stack()
    // %sp = input window's sp, %fp = input window's fp
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    .extern restore_all_cwps_from_stack
    _BSR    restore_all_cwps_from_stack, %g7
    NOP
    
    //ok, now we can resotre all registers of next thread
    WRCTL   %g0             ;Restore STATUS that we save in %g0
    NOP
    
    // let %sp = NSP
    MOV     %sp,%g6
    
    LDS     %g0, [%sp, 1]   ;Restore ISTATUS
    PFX     1
    WRCTL   %g0
    NOP
    
	LDS %g0, [%sp, 2]       ;Restore %g0~7  
	LDS %g1, [%sp, 3]
	LDS %g2, [%sp, 4]
	LDS %g3, [%sp, 5]
	LDS %g4, [%sp, 6]
	LDS %g5, [%sp, 7]
	LDS %g6, [%sp, 8]
	LDS %g7, [%sp, 9]
    
	LDS %o0, [%sp, 10]      ;Restore %o0~7  
	LDS %o1, [%sp, 11]
	LDS %o2, [%sp, 12]
	LDS %o3, [%sp, 13]
	LDS %o4, [%sp, 14]
	LDS %o5, [%sp, 15]
    // We will restore %o6(sp) at last.
	LDS %o7, [%sp, 17]
    
	LDS %l0, [%sp, 18]      ;Restore %l0~7  
	LDS %l1, [%sp, 19]
	LDS %l2, [%sp, 20]
	LDS %l3, [%sp, 21]
	LDS %l4, [%sp, 22]
	LDS %l5, [%sp, 23]
	LDS %l6, [%sp, 24]
	LDS %l7, [%sp, 25]
    
	LDS %i0, [%sp, 26]      ;Restore %i0~7  
	LDS %i1, [%sp, 27]
	LDS %i2, [%sp, 28]
	LDS %i3, [%sp, 29]
	LDS %i4, [%sp, 30]
	LDS %i5, [%sp, 31]
	LDS %i6, [%sp, 32]
	LDS %i7, [%sp, 33]

    LDS %o6, [%sp, 16]     ;Restore %sp 
    
	// Enable Interrupt
	PFX 9
	WRCTL %g0
	NOP
    
    // 'Next thread' already restore finish.
    // Execute it.
    JMP   %o7
    NOP
    
#------------------------------------------------------------------------------
# HAL longjmp, setjmp implementations
# hal_setjmp saves all registers into buffer supplied in %o0
	
	.globl	hal_setjmp
hal_setjmp:

    // Disable Interrupt here
    PFX     8
    WRCTL   %g0
    NOP

    // Save %l0 into buffer first, and let %lo = pointer of buffer  
    ST      [%o0], %l0
    MOV     %l0,%o0         
    
    // save all registers into buffer
    STP     [%l0 , 2], %g0          ;Save %g0~7
    STP     [%l0 , 3], %g1
    STP     [%l0 , 4], %g2
    STP     [%l0 , 5], %g3
    STP     [%l0 , 6], %g4
    STP     [%l0 , 7], %g5
    STP     [%l0 , 8], %g6
    STP     [%l0 , 9], %g7
    
    STP     [%l0 , 10], %o0         ;Save %o0~7
    STP     [%l0 , 11], %o1
    STP     [%l0 , 12], %o2
    STP     [%l0 , 13], %o3
    STP     [%l0 , 14], %o4
    STP     [%l0 , 15], %o5
    STP     [%l0 , 16], %o6
    STP     [%l0 , 17], %o7
    
    LD      %g0,[%o0]               ;Save %l0 that we put it in buffer[0] 
    STP     [%l0 , 18], %g0
    
    STP     [%l0 , 19], %l1         ;Save %l1~7
    STP     [%l0 , 20], %l2
    STP     [%l0 , 21], %l3
    STP     [%l0 , 22], %l4
    STP     [%l0 , 23], %l5
    STP     [%l0 , 24], %l6
    STP     [%l0 , 25], %l7
    
    STP     [%l0 , 26], %i0         ;Save %i0~7
    STP     [%l0 , 27], %i1
    STP     [%l0 , 28], %i2
    STP     [%l0 , 29], %i3
    STP     [%l0 , 30], %i4
    STP     [%l0 , 31], %i5
    PFX     %hi(32)
    STP     [%l0 , %lo(32)], %i6
    PFX     %hi(33)
    STP     [%l0 , %lo(33)], %i7
    
    PFX     1
    RDCTL   %g0
    STP     [%l0 , 1], %g0         ;Save ISTATUS

    RDCTL   %g0
    STP     [%l0 , 0], %g0         ;Save STATUS
    // now, %g0 = original STATUS

    MOV     %g1, %l0
    
    // backup_all_cwps_to_stack(), %sp = stack
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    .extern backup_all_cwps_to_stack
    _BSR    backup_all_cwps_to_stack, %g7
    NOP
    
    // let's return the original window
    WRCTL   %g0
    NOP
    
    // restore %sp, %l0
    MOV     %l0, %g1
    
    // Load ISTATUS from input buffer
    LDP     %g0, [%l0, 1]
    PFX     1
    WRCTL   %g0
    NOP

    // Load %g0~%g7 from input buffer
    LDP     %g0, [%l0, 2 ]
    LDP     %g1, [%l0, 3 ]
    LDP     %g2, [%l0, 4 ]
    LDP     %g3, [%l0, 5 ]
    LDP     %g4, [%l0, 6 ]
    LDP     %g5, [%l0, 7 ]
    LDP     %g6, [%l0, 8 ]
    LDP     %g7, [%l0, 9 ]    
    
    // Load %o1~%o7 from input buffer
    // Don't load %o0, because it's 'return value'
    LDP     %o1, [%l0, 11]
    LDP     %o2, [%l0, 12]
    LDP     %o3, [%l0, 13]
    LDP     %o4, [%l0, 14]
    LDP     %o5, [%l0, 15]
    LDP     %o6, [%l0, 16]    
    LDP     %o7, [%l0, 17]    
    
    // Load %l1~%l7 from input buffer, %l0 will load at last
    LDP     %l1, [%l0, 19]
    LDP     %l2, [%l0, 20]
    LDP     %l3, [%l0, 21]
    LDP     %l4, [%l0, 22]
    LDP     %l5, [%l0, 23]
    LDP     %l6, [%l0, 24]
    LDP     %l7, [%l0, 25]

    // Load %i07 from input buffer
    LDP     %i0, [%l0, 26]    
    LDP     %i1, [%l0, 27]
    LDP     %i2, [%l0, 28]
    LDP     %i3, [%l0, 29]
    LDP     %i4, [%l0, 30]
    LDP     %i5, [%l0, 31]
    PFX     %hi(32)
    LDP     %i6, [%l0, %lo(32)]
    PFX     %hi(33)
    LDP     %i7, [%l0, %lo(33)]
    
    // Load %l0
    LDP     %l0, [%l0, 18]
    
    // Enable Interrupt 
    PFX     9
    WRCTL   %g0
    NOP
    
    // set return value, and return
    MOVI    %o0,0
    JMP     %o7
    NOP
// end of hal_setjmp()    
    
//---------------------------------
//hal_longjmp:
//     %o0 = input buffer pointer
//     %o1 = return value

	.globl	hal_longjmp
hal_longjmp:

    // Disable Interrupt here
    PFX     8
    WRCTL   %g0
    NOP

    // Let %g1 = %o0 = input buffer pointer
    // %g2 = return value
    MOV     %g1, %o0
    MOV     %g2, %o1
    
    // Load STATUS first, and backup it to %g0
    LD      %g0, [%g1]
    WRCTL   %g0
    NOP
    
    // Load setjump()'s %sp
    PFX     16
    LD      %sp, [%g1]
    PFX     32
    LD      %fp, [%g1]

    // restore_all_cwps_from_stack()
    // %sp = input window's sp, %fp = input window's fp
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    .extern restore_all_cwps_from_stack
    _BSR    restore_all_cwps_from_stack, %g7
    NOP

    // get STATUS from %g0
    WRCTL   %g0
    NOP
    
    // now return the window of setjmp.
    // get the input buffer pointer from %g1
    MOV     %l0, %g1
    
    // Load ISTATUS from input buffer
    LDP     %g0, [%l0, 1]
    PFX     1
    WRCTL   %g0
    NOP

    // Put return value to %o0
    MOV     %o0, %g2
    
    // Load %g0~%g7 from input buffer
    LDP     %g0, [%l0, 2 ]
    LDP     %g1, [%l0, 3 ]
    LDP     %g2, [%l0, 4 ]
    LDP     %g3, [%l0, 5 ]
    LDP     %g4, [%l0, 6 ]
    LDP     %g5, [%l0, 7 ]
    LDP     %g6, [%l0, 8 ]
    LDP     %g7, [%l0, 9 ]    
    
    // Load %o1~%o7 from input buffer
    // Don't load %o0, because it's 'return value'
    LDP     %o1, [%l0, 11]
    LDP     %o2, [%l0, 12]
    LDP     %o3, [%l0, 13]
    LDP     %o4, [%l0, 14]
    LDP     %o5, [%l0, 15]
    LDP     %o6, [%l0, 16]    
    LDP     %o7, [%l0, 17]    
    
    // Load %l1~%l7 from input buffer, %l0 will load at last
    LDP     %l1, [%l0, 19]
    LDP     %l2, [%l0, 20]
    LDP     %l3, [%l0, 21]
    LDP     %l4, [%l0, 22]
    LDP     %l5, [%l0, 23]
    LDP     %l6, [%l0, 24]
    LDP     %l7, [%l0, 25]

    // Load %i07 from input buffer
    LDP     %i0, [%l0, 26]    
    LDP     %i1, [%l0, 27]
    LDP     %i2, [%l0, 28]
    LDP     %i3, [%l0, 29]
    LDP     %i4, [%l0, 30]
    LDP     %i5, [%l0, 31]
    PFX     %hi(32)
    LDP     %i6, [%l0, %lo(32)]
    PFX     %hi(33)
    LDP     %i7, [%l0, %lo(33)]
    
    // Load %l0
    LDP     %l0, [%l0, 18]

    // Enable Interrupt
    PFX     9
    WRCTL   %g0
    NOP
    
    // return the address that setting by hal_setjmp()
    JMP     %o7
    NOP
//end of hal_longjmp()    
#------------------------------------------------------------------------------
# end of context.S
