//=============================================================================
//
//	vectors.S
//
//	NIOS exception vectors
//
//=============================================================================
//###ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//###ECOSGPLCOPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):   Hao Tseng
// Contributors:    
// Date:        
// Purpose:	    NIOS exception vectors
// Description:	This file defines the code placed into the exception
//              vectors. It also contains the first level default VSRs
//		        that save and restore state for both exceptions and
//		        interrupts.
//
//####DESCRIPTIONEND####
//
//=============================================================================

#include <pkgconf/system.h>	
#include <pkgconf/hal.h>

#ifdef CYGPKG_KERNEL
# include <pkgconf/kernel.h>	
#endif

#include  CYGBLD_HAL_PLATFORM_H
#include <cyg/hal/arch.inc>	
#include <cyg/hal/nios_imacros.inc>

//-----------------------------------------------------------------------------
// (1) Reset vector

	// "ax": "a: section is allocatable" "x: section is executable"
	.section ".reset_vector","ax"
    .globl	reset_vector
reset_vector:
    PFX 8                       // First instruction is "disable the interrupt"
    WRCTL   %g0
    NOP
    MOVIA      %g0,_start@h     // Jump to _start
    JMP        %g0
    NOP

//-----------------------------------------------------------------------------
// (2) Execption vectors
//
//(2)-1 CWP underflow/overflow management
//      CWPUnderflowTrapHandler:
//      CWPOverflowTrapHandler:
    
	.section ".bss" 
cwp_backup:         ; this is temporary storage when CWP underflow/overflow
	.rept 32        ; management
	.long 0
	.endr
                    
 	.section ".text","ax"
CWPUnderflowTrapHandler:
// Now, CWP = LO_LIMIT - 1
// %sp = value after SAVE instruction
// %o7 = return address
// IE = 0
// ISTATUS = STATUS before interrupt 

    // First , we backup %sp, %o7, %g, %i to cwp_backup
    MOVIA   %l0, cwp_backup
    STP     [%l0, 0], %sp       // backup %sp
    STP     [%l0, 1], %o7       // backup %o7
    STP     [%l0, 2], %g0       // backup %g
    STP     [%l0, 3], %g1
    STP     [%l0, 4], %g2
    STP     [%l0, 5], %g3
    STP     [%l0, 6], %g4
    STP     [%l0, 7], %g5
    STP     [%l0, 8], %g6
    STP     [%l0, 9], %g7
    STP     [%l0, 10],%i0       // backup %i
    STP     [%l0, 11],%i1
    STP     [%l0, 12],%i2
    STP     [%l0, 13],%i3
    STP     [%l0, 14],%i4
    STP     [%l0, 15],%i5
    STP     [%l0, 16],%i6
    STP     [%l0, 17],%i7

    //<<Hao>> Debugging ==>    
//    MOVIA   %l0, 0x800    
//lp1:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	lp1			; loop until empty
//    NOP
//    // Display 'v'
//    MOVIP   %l1, 0x76767676
//    STP     [%l0,1], %l1
//
//lp2:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	lp2			; loop until empty
//    NOP
//    
//    // Display CWP + '0'    
//    RDCTL   %l1
//    ANDIP   %l1, 0x1f0    
//    LSRI    %l1,4
//    ADDIP   %l1,0x30
//    STP     [%l0,1], %l1
//    
//lp3:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	lp3			; loop until empty
//    NOP
//    
//    MOVIP   %l1, 0x0d0d0d0d
//    STP     [%l0,1], %l1
//
//lp4:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	lp4			; loop until empty
//    NOP
//
//    MOVIP   %l1, 0x0a0a0a0a
//    STP     [%l0,1], %l1
//    
//lp5:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	lp5			; loop until empty
//    NOP
//    
//    
//<==

    // backup_all_cwps_to_stack(), %sp = stack
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    _BSR    backup_all_cwps_to_stack, %g7
    NOP

    // now all cwp regs already save to stack

    // Set current CWP = HI_LIMIT
    PFX     2
    RDCTL   %g0
    LSRI    %g0,1
    ANDIP   %g0, 0x1f0          // %g0 = HI_LIMIT in cwp position
    RDCTL   %g1
    ANDNIP  %g1, 0x1f0
    OR      %g1,%g0
    WRCTL   %g1                 // now, CWP is at HI_LIMIT
    NOP
    
    // Set ISTATUS's CWP = HI_LIMIT
    PFX     1
    RDCTL   %g1
    ANDNIP  %g1, 0x1f0
    OR      %g1,%g0
    PFX     1
    WRCTL   %g1                 // now, ISTATUS's CWP is at HI_LIMIT
    NOP
    
    // Get %sp, %o7, %g, %i from  cwp_backup
    MOVIA   %l0, cwp_backup
    LDP     %sp, [%l0, 0]       // restore %sp
    LDP     %o7, [%l0, 1]       // restore %o7
    LDP     %g0, [%l0, 2]       // restore %g
    LDP     %g1, [%l0, 3]
    LDP     %g2, [%l0, 4]
    LDP     %g3, [%l0, 5]
    LDP     %g4, [%l0, 6]
    LDP     %g5, [%l0, 7]
    LDP     %g6, [%l0, 8]
    LDP     %g7, [%l0, 9] 
    LDP     %i0, [%l0, 10]      // restore %i
    LDP     %i1, [%l0, 11]
    LDP     %i2, [%l0, 12]
    LDP     %i3, [%l0, 13]
    LDP     %i4, [%l0, 14]
    LDP     %i5, [%l0, 15]
    LDP     %i6, [%l0, 16]
    LDP     %i7, [%l0, 17]    
    
    // return    
	TRET	%o7
    NOP


CWPOverflowTrapHandler:
// Now, CWP = HI_LIMIT
// %sp = value before RESTORE instruction
// %o7 = return address
// IE = 0
// ISTATUS = STATUS before interrupt 

    // First , we backup %sp, %o7, %g, %i to cwp_backup
    MOVIA   %l0, cwp_backup
    STP     [%l0, 0], %sp       // backup %sp
    STP     [%l0, 1], %o7       // backup %o7
    STP     [%l0, 2], %g0       // backup %g
    STP     [%l0, 3], %g1
    STP     [%l0, 4], %g2
    STP     [%l0, 5], %g3
    STP     [%l0, 6], %g4
    STP     [%l0, 7], %g5
    STP     [%l0, 8], %g6
    STP     [%l0, 9], %g7
    STP     [%l0, 10],%i0       // backup %i
    STP     [%l0, 11],%i1
    STP     [%l0, 12],%i2
    STP     [%l0, 13],%i3
    STP     [%l0, 14],%i4
    STP     [%l0, 15],%i5
    STP     [%l0, 16],%i6
    STP     [%l0, 17],%i7


    //<<Hao>> Debugging ==>    
//    MOVIA   %l0, 0x800    
//yp1:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	yp1			; loop until empty
//    NOP
//    
//    // Display '^'
//    MOVIP   %l1, 0x5e5e5e5e
//    STP     [%l0,1], %l1
//
//yp2:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	yp2			; loop until empty
//    NOP
//
//    // Display CWP + '0'    
//    RDCTL   %l1
//    ANDIP   %l1, 0x1f0    
//    LSRI    %l1,4
//    ADDIP   %l1,0x30
//    STP     [%l0,1], %l1
//    
//yp3:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	yp3			; loop until empty
//    NOP
//    
//    
//    MOVIP   %l1, 0x0d0d0d0d
//    STP     [%l0,1], %l1
//
//yp4:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	yp4			; loop until empty
//    NOP
//
//    MOVIP   %l1, 0x0a0a0a0a
//    STP     [%l0,1], %l1
//    
//yp5:    
//    LDP     %l1,[%l0, 2]
//	SKP1	%l1, 6	; check the ready bit
//	BR	yp5			; loop until empty
//    NOP
//    
//<==

    
    // Set current CWP = LO_LIMIT - 1, ISTATUS's CWP = LO_LIMIT
    MOVIP   %g2, 0x1f0          // Set %g2 = mask of CWP
    PFX     2
    RDCTL   %g0
    LSLI    %g0,4
    AND     %g0, %g2            // %g0 = LO_LIMIT in cwp position
    PFX     1
    RDCTL   %g1
    ANDN    %g1, %g2
    OR      %g1, %g0
    PFX     1
    WRCTL   %g1                 // now, ISTATUS's CWP is at LO_LIMIT
    NOP
    RDCTL   %g1
    ANDN    %g1, %g2
    OR      %g1, %g0
    SUBIP   %g1, 0x10 
    WRCTL   %g1                 // now, current CWP is at (LO_LIMIT -1)
    NOP

    // get sp value from  cwp_backup
    MOVIA   %sp, cwp_backup
    PFX     16
    LD      %fp, [%sp]          // Let %fp = Window(LO_LIMIT -1)'s fp
    LD      %sp, [%sp]          // Let %sp = Window(LO_LIMIT -1)'s sp

    // restore_all_cwps_from_stack()
    // %sp = input window's sp, %fp = input window's fp
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    _BSR    restore_all_cwps_from_stack, %g7
    NOP
    
    // Set current CWP = LO_LIMIT - 1 again 
    WRCTL   %g1
    NOP

    // Get %sp, %o7, %g, %i from  cwp_backup
    MOVIA   %l0, cwp_backup
    LDP     %sp, [%l0, 0]       // restore %sp
    LDP     %o7, [%l0, 1]       // restore %o7
    LDP     %g0, [%l0, 2]       // restore %g
    LDP     %g1, [%l0, 3]
    LDP     %g2, [%l0, 4]
    LDP     %g3, [%l0, 5]
    LDP     %g4, [%l0, 6]
    LDP     %g5, [%l0, 7]
    LDP     %g6, [%l0, 8]
    LDP     %g7, [%l0, 9] 
    LDP     %i0, [%l0, 10]      // restore %i
    LDP     %i1, [%l0, 11]
    LDP     %i2, [%l0, 12]
    LDP     %i3, [%l0, 13]
    LDP     %i4, [%l0, 14]
    LDP     %i5, [%l0, 15]
    LDP     %i6, [%l0, 16]
    LDP     %i7, [%l0, 17]    
    
    // return    
	TRET	%o7
    NOP


//
//(2)-2 Null interrupt handler
//	
NullInterruptHandler:
	TRET    %o7            // just return back
    NOP

//
//(2)-3 Exception VSR handler
ExceptionVSRHandler:
	// IE cleared here

    //======================================================
    //(1)Save CPU State
    // We create a space in interruptee's stack for HAL_SavedRegisters, return address
    //   %fp(thread'sp):  xxxx   
    //                    yyyy  <-- Return Address(4 bytes)
    //                    zzzz            
    //                     |
    //                     |    <-- HAL_SavedRegisters (SAVED_REGS_SIZE)
    //                     |
    //             NSP--> zzzz
    //
    MOV     %l0,%fp
    SUBIP   %l0,SAVED_REGS_SIZE + 4
    // Now %l0 is the pointer of HAL_SavedRegisters = NSP
    
    // Save the globals to HAL_SavedRegisters
	STP	[%l0,2],%g0
	STP	[%l0,3],%g1
	STP	[%l0,4],%g2
	STP	[%l0,5],%g3
	STP	[%l0,6],%g4
	STP	[%l0,7],%g5
	STP	[%l0,8],%g6
	STP	[%l0,9],%g7
    
    // Save the interruptee's %o0~7 to HAL_SavedRegisters(now is at %i0~7)
	STP	[%l0,10],%i0
	STP	[%l0,11],%i1
	STP	[%l0,12],%i2
	STP	[%l0,13],%i3
	STP	[%l0,14],%i4
	STP	[%l0,15],%i5
	STP	[%l0,16],%i6
	STP	[%l0,17],%i7

	// Save the return address to the interruptee, across the RESTORE
    PFX %hi(34)
	STP	[%l0, %lo(34)], %o7
    
    mov %g0,%l0         // let %g0 is the pointer of HAL_SavedRegisters = NSP
    
	RESTORE
	// CWP is now in the interruptee's level.
    
    mov %sp, %g0        //original sp already be saved , so we can chagne sp value = NSP

	// save %l0~7 to HAL_SavedRegisters
	STS	[%sp, 18], %l0
	STS	[%sp, 19], %l1
	STS	[%sp, 20], %l2
	STS	[%sp, 21], %l3
	STS	[%sp, 22], %l4
	STS	[%sp, 23], %l5
	STS	[%sp, 24], %l6
	STS	[%sp, 25], %l7
	// save %i0~7
	STS	[%sp, 26], %i0
	STS	[%sp, 27], %i1
	STS	[%sp, 28], %i2
	STS	[%sp, 29], %i3
	STS	[%sp, 30], %i4
	STS	[%sp, 31], %i5
	STS	[%sp, 32], %i6
	STS	[%sp, 33], %i7
    
	// save STATUS
	PFX	0
	RDCTL	%g1
	STS	[%sp,0], %g1

	// save ISTATUS
	PFX	1
	RDCTL	%g1
	STS	[%sp,1], %g1

    // until now %sp = NSP
  
    //======================================================
    //(4)Switch to an interrupt stack or using thread's stack
    
    // Save NSP to %g6
    MOV %g6,%sp
    
    // backup_all_cwps_to_stack(), %sp = stack
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    ADDIP   %sp, SAVED_REGS_SIZE + 4    
    _BSR    backup_all_cwps_to_stack, %g7
    NOP

    // Let CWP up to HI_LIMIT
	PFX	2
	RDCTL	%g1
	LSRI	%g1,1
	ANDIP	%g1,0x01f0		// g1 = HI_LIMIT in position for ctl0
	RDCTL	%g0
	ANDNIP	%g0,0x81f0		// IRQ disabled, and space for cwp
	OR	%g0,%g1
	WRCTL	%g0
	NOP

    // Let %sp = NSP
    MOV     %sp,%g6
    
#if defined(CYGSEM_HAL_ROM_MONITOR) && \
    defined(CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK)
    //======================================================
    // Using interrupt stack

	// get interrupt stack address into %l0 (top -- High)
	MOVIA %g0, cyg_interrupt_stack

	// get the address of cyg_interrupt_stack_base into %l0 --> %g3 
	// (bottom -- low)
	MOVIA %g1, cyg_interrupt_stack_base	

	CMP	%g6, %g1	// below stack base?
	IFS	cc_mi		// execute next instruction if below(not on the IRQ stack yet), switch needed
	BR _VSR1   	// if so, switch.
    NOP

	CMP	%g6, %g0	// below top? (DELAY SLOT) then no switch needed
	IFS	cc_mi		// execute next instruction if below means already on the interrupt stack, switch not needed
	BR  _VSR2
	NOP				// (DELAY SLOT)

_VSR1:	// switch to the interrupt stack
    MOV %sp,%g0
    
_VSR2:	
	  // continue as before, if already on the interrupt stack
#endif

    MOV     %l0, %g6            // Let %l0 = NSP

    // Until now, we already do something below
    //Case (Using interrupt stack): (1) Save all regs into thread's stack, 
    //                              (2) Save all Window into thread's into stack    
    //                              (3) change %sp to interrupt stack
    //                              (4) Set CWP to HI_LIMIT
    //                              (5) %l0 = NSP    
    //Case (Using thread's stack):  (1) Save all regs into thread's stack,
    //                              (2) Save all Window into thread's into stack
    //                              (3) Modify %sp = NSP
    //                              (4) Set CWP to HI_LIMIT
    //                              (5) %l0 = NSP

    //======================================================
    //(5) Decode the actual exception number
    // let %o0 = vector number =  IPRI - CYGNUM_HAL_EXCEPTION_MIN
    RDCTL   %l1                 //read STATUS register
    ANDIP   %l1,0x7e00          
    LSRI    %l1,9               
    SUBIP   %l1, CYGNUM_HAL_EXCEPTION_MIN  //let %l1=IPRI - CYGNUM_HAL_EXCEPTION_MIN
    MOV     %o0,%l1
    
    //======================================================
    //(6)Get VSR handler address 
    // Get hal_vsr_table[vsr_vector_num]
    LSLI    %l1,2                       //%l1 = vsr_vector_num << 2
    .extern hal_vsr_table
    MOVIA   %g1, hal_vsr_table
    ADD     %g1,%l1                   
    LD      %g1,[%g1]                   //now %g1 = *hal_vsr_table[vsr_vector_num]

    //======================================================
    //(7) Re-enable interrupt to let CWP management can be worked
	PFX	9
	WRCTL	%g0				// (value doesn't matter)
    NOP
    
    //======================================================
    //(8)Call the ISR
    
    MOV     %o1, %l0        //%o1 = pointer of regs
                            //%o0 = vsr vector number
	CALL %g1                //VSR(vector, *Regs)
	NOP

    //======================================================
    //(10)return process

    //Disable all interrupt
    PFX     8
    WRCTL   %g0
    NOP                    // (delay slot)   
    
    // backup NSP value to %g6
    MOV     %g6, %l0 
    
    // Restore STATUS first, and let %g0 = STATUS
    LD      %g0, [%g6]          ; restore STATUS from stack
    WRCTL   %g0                 ; write STATUS value
    NOP

    //======================================================
    // Restore all CWP data from stack
    //
    // restore_all_cwps_from_stack()
    // %sp = input window's sp, %fp = input window's fp    
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    PFX     32
    LD      %fp,[%g6]
    MOV     %sp,%g6
    ADDIP   %sp, SAVED_REGS_SIZE + 4    
    _BSR    restore_all_cwps_from_stack, %g7
    NOP

    // Restore STATUS first, and let %g0 = STATUS
    LD      %g0, [%g6]          ; restore STATUS from NSP
    WRCTL   %g0                 ; write STATUS value
    NOP

    //take back NSP value
    MOV     %sp,%g6

    // Unitl now, STATUS already restored & %sp = NSP

    //Resotre all registers that stored at HAL_SavedRegisters
    LDS     %l0, [%sp , 18]                  //restore %l0~7
    LDS     %l1, [%sp , 19]
    LDS     %l2, [%sp , 20]
    LDS     %l3, [%sp , 21]
    LDS     %l4, [%sp , 22]
    LDS     %l5, [%sp , 23]
    LDS     %l6, [%sp , 24]
    LDS     %l7, [%sp , 25]    
    LDS     %i0, [%sp , 26]                  //restore %i0~7
    LDS     %i1, [%sp , 27]
    LDS     %i2, [%sp , 28]
    LDS     %i3, [%sp , 29]
    LDS     %i4, [%sp , 30]
    LDS     %i5, [%sp , 31]
    LDS     %i6, [%sp , 32]
    LDS     %i7, [%sp , 33]    
    
    // let CWP - 1
    RDCTL   %g0
    SUBIP   %g0,0x0010
    WRCTL   %g0
    NOP    
    
    MOV     %sp,%fp
    //After CWP -= 1, The interruptee's %o0~o7 now is at %i0~7
    LDS     %i0, [%sp , 10]                  //restore %o0~7
    LDS     %i1, [%sp , 11]
    LDS     %i2, [%sp , 12]
    LDS     %i3, [%sp , 13]
    LDS     %i4, [%sp , 14]
    LDS     %i5, [%sp , 15]
    LDS     %i6, [%sp , 16]
    LDS     %i7, [%sp , 17]
    
    LDS     %g0, [%sp , 1]         //restore ISTATUS
    PFX     1
    WRCTL   %g0
    NOP

    LDS     %o7, [%sp , 34]         //get return address

    LDS     %g0, [%sp , 2]                  //restore %g0~7
    LDS     %g1, [%sp , 3]
    LDS     %g2, [%sp , 4]
    LDS     %g3, [%sp , 5]
    LDS     %g4, [%sp , 6]
    LDS     %g5, [%sp , 7]
    LDS     %g6, [%sp , 8]
    LDS     %g7, [%sp , 9]
    
    TRET    %o7     //return
    NOP     
//ExceptionVSRHandler finish.


//
//(2)-4 Interrupt VSR handler
InterruptVSRHandler:

	// IE cleared here

    //======================================================
    //(1)Save CPU State
    // We create a space in interruptee's stack for HAL_SavedRegisters, return address
    //   %fp(thread'sp):  xxxx   
    //                    yyyy  <-- Return Address(4 bytes)
    //                    zzzz            
    //                     |
    //                     |    <-- HAL_SavedRegisters (SAVED_REGS_SIZE)
    //                     |
    //             NSP--> zzzz
    //
    MOV     %l0,%fp
    SUBIP   %l0,SAVED_REGS_SIZE + 4
    // Now %l0 is the pointer of HAL_SavedRegisters = NSP
    
    // Save the globals to HAL_SavedRegisters
	STP	[%l0,2],%g0
	STP	[%l0,3],%g1
	STP	[%l0,4],%g2
	STP	[%l0,5],%g3
	STP	[%l0,6],%g4
	STP	[%l0,7],%g5
	STP	[%l0,8],%g6
	STP	[%l0,9],%g7
    
    // Save the interruptee's %o0~7 to HAL_SavedRegisters(now is at %i0~7)
	STP	[%l0,10],%i0
	STP	[%l0,11],%i1
	STP	[%l0,12],%i2
	STP	[%l0,13],%i3
	STP	[%l0,14],%i4
	STP	[%l0,15],%i5
	STP	[%l0,16],%i6
	STP	[%l0,17],%i7

	// Save the return address to the interruptee, across the RESTORE
    PFX %hi(34)
	STP	[%l0, %lo(34)], %o7
    
    mov %g0,%l0         // let %g0 is the pointer of HAL_SavedRegisters = NSP
    
	RESTORE
	// CWP is now in the interruptee's level.
    
    mov %sp, %g0        //original sp already be saved , so we can chagne sp value = NSP

	// save %l0~7 to HAL_SavedRegisters
	STS	[%sp, 18], %l0
	STS	[%sp, 19], %l1
	STS	[%sp, 20], %l2
	STS	[%sp, 21], %l3
	STS	[%sp, 22], %l4
	STS	[%sp, 23], %l5
	STS	[%sp, 24], %l6
	STS	[%sp, 25], %l7
	// save %i0~7
	STS	[%sp, 26], %i0
	STS	[%sp, 27], %i1
	STS	[%sp, 28], %i2
	STS	[%sp, 29], %i3
	STS	[%sp, 30], %i4
	STS	[%sp, 31], %i5
	STS	[%sp, 32], %i6
	STS	[%sp, 33], %i7
    
	// save STATUS
	PFX	0
	RDCTL	%g1
	STS	[%sp,0], %g1

	// save ISTATUS
	PFX	1
	RDCTL	%g1
	STS	[%sp,1], %g1

    // until now %sp = NSP
    
    //======================================================
    //(2)Platform interrupt pre_process
    //    
    .extern hal_platform_intr_pre_process
    _BSR    hal_platform_intr_pre_process,%g7
    NOP
    
    //======================================================
    //(3)Increment the kernel scheduler lock
    //    
#ifdef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
	// Lock the scheduler
	.extern	cyg_scheduler_sched_lock
	MOVIA %g0, cyg_scheduler_sched_lock
	// load scheduler counter into %g1
	LD	%g1, [ %g0 ]
	// lock scheduler
	ADDI %g1, 1
	// saved back
	ST	[ %g0 ], %g1
#endif
    
    //======================================================
    //(4)Switch to an interrupt stack or using thread's stack
    
    // Save NSP to %g6
    MOV %g6,%sp
    
    
    //======================================================
    // Backup all CWP data into stack
    //
    // backup_all_cwps_to_stack(), %sp = stack
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    ADDIP   %sp, SAVED_REGS_SIZE + 4    
    _BSR    backup_all_cwps_to_stack, %g7
    NOP

    // Let CWP up to HI_LIMIT
	PFX	2
	RDCTL	%g1
	LSRI	%g1,1
	ANDIP	%g1,0x01f0		// g1 = HI_LIMIT in position for ctl0
	RDCTL	%g0
	ANDNIP	%g0,0x81f0		// IRQ disabled, and space for cwp
	OR	%g0,%g1
	WRCTL	%g0
	NOP

    // Let %sp = NSP
    MOV     %sp,%g6
    
#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
    //======================================================
    // Using interrupt stack

	// get interrupt stack address into %l0 (top -- High)
	MOVIA %g0, cyg_interrupt_stack

	// get the address of cyg_interrupt_stack_base into %l0 --> %g3 
	// (bottom -- low)
	MOVIA %g1, cyg_interrupt_stack_base	

	CMP	%g6, %g1	// below stack base?
	IFS	cc_mi		// execute next instruction if below(not on the IRQ stack yet), switch needed
	BR _EISR1   	// if so, switch.
    NOP

	CMP	%g6, %g0	// below top? (DELAY SLOT) then no switch needed
	IFS	cc_mi		// execute next instruction if below means already on the interrupt stack, switch not needed
	BR  _EISR2
	NOP				// (DELAY SLOT)

_EISR1:	// switch to the interrupt stack
    MOV %sp,%g0
    
_EISR2:	
	  // continue as before, if already on the interrupt stack
#endif

    MOV     %l0, %g6            // Let %l0 = NSP

    // Until now, we already do something below
    //Case (Using interrupt stack): (1) Save all regs into thread's stack, 
    //                              (2) Save all Window into thread's into stack    
    //                              (3) change %sp to interrupt stack
    //                              (4) Set CWP to HI_LIMIT
    //                              (5) %l0 = NSP    
    //Case (Using thread's stack):  (1) Save all regs into thread's stack,
    //                              (2) Save all Window into thread's into stack
    //                              (3) Modify %sp = NSP
    //                              (4) Set CWP to HI_LIMIT
    //                              (5) %l0 = NSP

    //======================================================
    //(5) Decode the actual interrupt number
    // let %o0 = vector number =  IPRI - CYGNUM_HAL_ISR_MIN
    RDCTL   %l1                 //read STATUS register
    ANDIP   %l1,0x7e00          
    LSRI    %l1,9               
    SUBIP   %l1, CYGNUM_HAL_ISR_MIN     //let %l1=IPRI - CYGNUM_HAL_ISR_MIN
    MOV     %o0,%l1
    
    //======================================================
    //(6)Get ISR pointer and it's data pointer
    // Get hal_interrupt_data[isr_vector_num]
    LSLI    %l1,2                       //%l1 = isr_vector_num << 2
    .extern hal_interrupt_data
    MOVIA   %g1, hal_interrupt_data
    ADD     %g1,%l1                   
    LD      %o1,[%g1]                   //now, %o1 = hal_interrupt_data[isr_vector_num]
    // Get *hal_interrupt_handlers[isr_vector_num]
    .extern hal_interrupt_handlers
    MOVIA   %g1, hal_interrupt_handlers
    ADD     %g1,%l1
    LD      %g1,[%g1]                   //now %g1 = *hal_interrupt_[isr_vector_num]
    
    //======================================================
    //(7) Re-enable interrupt to permit nesting
	PFX	9
	WRCTL	%g0				// (value doesn't matter)
    NOP
    
    //======================================================
    //(8)Call the ISR
	CALL %g1                // ISR(vector, hal_interrupt_data[ vector ])
	NOP
         
    
#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
	// If we are returning from the last nested interrupt, move back
	// to the thread stack. interrupt_end() must be called on the
	// thread stack since it potentially causes a context switch.
    
    //switch %sp back to NSP
    MOV %sp,%l0
#endif

#ifdef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
    //======================================================
    //(9)Call interrupt_end(r, cyg_hal_interrupt_objects[vector], regs)
	// - r is the return value from the ISR
	// - regs points to saved CPU context
    
    .extern hal_interrupt_objects
    MOVIA   %g1, hal_interrupt_objects
    ADD     %g1,%l1                   
    LD      %o1,[%g1]                   //now, %o1 = hal_interrupt_objects[vector]
    MOV     %o2,%sp                     // %o2 = NSP = pointer of regs
                                        // %o0 = ISR() return value
  	.extern	interrupt_end
    _BSR    interrupt_end,%g7           // Call interrupt_end()
    NOP    
#endif    

    //======================================================
    //(10)return process

    //Disable all interrupt
    PFX     8
    WRCTL   %g0
    NOP                    // (delay slot)
    
    // backup NSP value to %g6
    MOV     %g6, %sp 
    
    // Restore STATUS first, and let %g0 = STATUS
    LDS     %g0, [%sp, 0]       ; restore STATUS from stack
    WRCTL   %g0                 ; write STATUS value
    NOP

    //======================================================
    // Restore all CWP data from stack
    //
    // restore_all_cwps_from_stack()
    // %sp = input window's sp, %fp = input window's fp
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    PFX     32
    LD      %fp,[%g6]
    MOV     %sp,%g6
    ADDIP   %sp, SAVED_REGS_SIZE + 4    
    _BSR    restore_all_cwps_from_stack, %g7
    NOP

    // Restore STATUS first, and let %g0 = STATUS
    LD      %g0, [%g6]          ; restore STATUS from NSP
    WRCTL   %g0                 ; write STATUS value
    NOP

    //take back NSP value
    MOV     %sp,%g6

    // Unitl now, STATUS already restored & %sp = NSP

    //Resotre all registers that stored at HAL_SavedRegisters
    LDS     %l0, [%sp , 18]                  //restore %l0~7
    LDS     %l1, [%sp , 19]
    LDS     %l2, [%sp , 20]
    LDS     %l3, [%sp , 21]
    LDS     %l4, [%sp , 22]
    LDS     %l5, [%sp , 23]
    LDS     %l6, [%sp , 24]
    LDS     %l7, [%sp , 25]    
    LDS     %i0, [%sp , 26]                  //restore %i0~7
    LDS     %i1, [%sp , 27]
    LDS     %i2, [%sp , 28]
    LDS     %i3, [%sp , 29]
    LDS     %i4, [%sp , 30]
    LDS     %i5, [%sp , 31]
    LDS     %i6, [%sp , 32]
    LDS     %i7, [%sp , 33]    
    
    // let CWP - 1
    RDCTL   %g0
    SUBIP   %g0,0x0010
    WRCTL   %g0
    NOP    
    
    MOV     %sp,%fp
    //After CWP -= 1, The interruptee's %o0~o7 now is at %i0~7
    LDS     %i0, [%sp , 10]                  //restore %o0~7
    LDS     %i1, [%sp , 11]
    LDS     %i2, [%sp , 12]
    LDS     %i3, [%sp , 13]
    LDS     %i4, [%sp , 14]
    LDS     %i5, [%sp , 15]
    LDS     %i6, [%sp , 16]
    LDS     %i7, [%sp , 17]
    
    LDS     %g0, [%sp , 1]         //restore ISTATUS
    PFX     1
    WRCTL   %g0
    NOP

    LDS     %o7, [%sp , 34]         //get return address

    LDS     %g0, [%sp , 2]                  //restore %g0~7
    LDS     %g1, [%sp , 3]
    LDS     %g2, [%sp , 4]
    LDS     %g3, [%sp , 5]
    LDS     %g4, [%sp , 6]
    LDS     %g5, [%sp , 7]
    LDS     %g6, [%sp , 8]
    LDS     %g7, [%sp , 9]
    
    TRET    %o7     //return
    NOP     
//InterruptVSRHandler finish.



//
//(2)-5 hal_interrupt_stack_call_pending_DSRs

// only applys to the interrupt stack used
#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK

// hal_interrupt_stack_call_pending_DSRs will **not** be called in nested interrupt coz unlock_inner( ) will not called
// which will in turns call Cyg_Interrupt::call_pending_DSRs( ) -->  HAL_INTERRUPT_STACK_CALL_PENDING_DSRS( ) 
// --> hal_interrupt_stack_call_pending_DSRs( ) (this routine)

// This routine can **only** be called from a thread stack
// It switches to the interrupt stack then calls back to the
// kernel to execute DSRs.

	.global hal_interrupt_stack_call_pending_DSRs
hal_interrupt_stack_call_pending_DSRs:
	SAVE %sp, -0x17

    //Disable Interrupt
    PFX     8
    WRCTL   %g0
    NOP

    // let %l0 = NSP = %sp - (8+1) * 4
    MOV     %l0, %sp
    SUBIP   %l0, 36
    
    // Save STATUS & %i into *(NSP)
    RDCTL   %g0
    STP     [%l0,0], %g0
    STP     [%l0,1], %i0
    STP     [%l0,2], %i1
    STP     [%l0,3], %i2
    STP     [%l0,4], %i3
    STP     [%l0,5], %i4
    STP     [%l0,6], %i5
    STP     [%l0,7], %i6
    STP     [%l0,8], %i7

    // Let %g6 = NSP    
    MOV     %g6,%l0

    // backup_all_cwps_to_stack(), %sp = stack
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    .extern backup_all_cwps_to_stack
    _BSR    backup_all_cwps_to_stack, %g7
    NOP
    
    //Set current CWP to HI_LIMIT
    PFX     2
    RDCTL   %g1
    LSRI    %g1, 1
    ANDIP   %g1, 0x1f0
    ANDNIP  %g0, 0x1f0
    OR      %g0,%g1                 
    WRCTL   %g0                     // Now current CWP is at HI_LIMIT
    NOP
    
    // Let %l0 = NSP
    MOV     %l0,%g6 
    
    // switch stack to interrupt stack
	MOVIA   %sp, cyg_interrupt_stack
	
    // Enable Interrupt
    PFX     9
    WRCTL   %g0
    NOP
	// call cyg_interrupt_call_pending_DSRs
	.extern cyg_interrupt_call_pending_DSRs
	MOVIA %l1, cyg_interrupt_call_pending_DSRs@h
	CALL  %l1
	NOP

    //Disable Interrupt
    PFX     8
    WRCTL   %g0
    NOP

    // restore STATUS from *(NSP)
    MOV     %g6, %l0            // Let %g6 = NSP
    LD      %g0, [%g6]
    WRCTL   %g0
    NOP

    // restore CWP data
    MOV     %sp, %g6
    LDS     %fp,[%sp, 7]
    ADDIP   %sp, 36
    
    // restore_all_cwps_from_stack(), 
    // %sp = input window's sp, %fp = input window's fp
    // This function use %g3,%g4,%g5,%g7 register
    // And it maybe destory status & %sp registers.
    .extern restore_all_cwps_from_stack
    _BSR    restore_all_cwps_from_stack, %g7
    NOP
    
    // set back STATUS
    WRCTL   %g0
    NOP
    
    // set back %i, %sp
    MOV     %l0,%g6
    LDP     %i0, [%l0, 1]
    LDP     %i1, [%l0, 2]
    LDP     %i2, [%l0, 3]
    LDP     %i3, [%l0, 4]
    LDP     %i4, [%l0, 5]
    LDP     %i5, [%l0, 6]
    LDP     %i6, [%l0, 7]
    LDP     %i7, [%l0, 8]

    MOV     %sp, %g6
    ADDIP   %sp, 36

    // Enable Interrupt
    PFX     9
    WRCTL   %g0
    NOP
	
	RET 
	RESTORE

#endif // CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK



//-----------------------------------------------------------------------------
// (3) Define startup stack & interrupt stack
//
// cyg_interrupt_stack			|	"*high* memory address"
//								|	
//								|
//								|
//								|
// cyg_interrupt_stack_base     V	"*low* memory address"

// Note: stack grows downward
// we define the startup stack


#ifndef CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
#define CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE 4096
#endif

   	.section ".bss"
	.balign 4
	.global cyg_interrupt_stack_base
cyg_interrupt_stack_base:
	
	.rept CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
	.word 0
	.endr

	.balign 4
	.global cyg_interrupt_stack
cyg_interrupt_stack:
	.rept 64
	.word 0
	.endr

//-----------------------------------------------------------------------------
// (4) Kernel startup code
//
   	.section ".text","ax"
_start:
	// Preset CWP up to top
	PFX	2
	RDCTL	%g1
	LSRI	%g1,1
	ANDIP	%g1,0x01f0		// g1 = HI_LIMIT in position for ctl0
	RDCTL	%g0
	ANDNIP	%g0,0x81f0		// IRQ disabled, and space for cwp
	OR	%g0,%g1
	WRCTL	%g0
	NOP

    
    // Reset platfrom & initial basic hardware such as DRAM, cache
    .extern	hal_platform_reset
    _BSR     hal_platform_reset,%g7
    NOP

#ifndef CYGSEM_HAL_USE_ROM_MONITOR    
    // Work as ROM Monitor
    
    // The Nios architecture do not have exceptoion except reset(#0)
    // CWP management(#1,#2).
    // But we reserve some vectors for implement eCos exception, if defined CYGHWR_NIOS_SUPPORT_EXCEPTION
    // Other vectors are used for interrupt.
    
    // first set reset, CWP underflow/overflow handler
    MOVIA   %l0, CYGHWR_NIOS_VECTOR_TABLE_ADDRESS
    MOVIA   %g0,reset_vector@h                //set reset handler
    ST      [%l0],%g0
    ADDI    %l0,4
    MOVIA   %g0,CWPUnderflowTrapHandler@h     // set CWP underflow handler
    ST      [%l0],%g0
    ADDI    %l0,4
    MOVIA   %g0,CWPOverflowTrapHandler@h      // set CWP overflow handler
    ST      [%l0],%g0
    ADDI    %l0,4

    // set exception vectors
    MOVIA   %g0,ExceptionVSRHandler@h    // set exception vectors to ExceptionVSRHandler
    MOVIA   %g2,CYGNUM_HAL_EXCEPTION_COUNT    
_SetVectorLoop:                             
    ST      [%l0],%g0
    ADDI    %l0,4
    SUBI    %g2,1
    IFRNZ  %g2
    BR      _SetVectorLoop
    NOP                     // (delay slot)    

#endif //#ifndef CYGSEM_HAL_USE_ROM_MONITOR    
    MOVIA   %l1, CYGNUM_HAL_ISR_MIN
    LSLI    %l1, 2
    MOVIA   %l0, CYGHWR_NIOS_VECTOR_TABLE_ADDRESS
    ADD     %l0, %l1       
    
    // set interruupt vectors
    MOVIA   %g0,InterruptVSRHandler@h    // set others to InterrupterVSRHandler
    MOVIA   %g2,CYGNUM_HAL_ISR_COUNT    
_SetVectorLoop2:                             
    ST      [%l0],%g0
    ADDI    %l0,4
    SUBI    %g2,1
    IFRNZ   %g2
    BR      _SetVectorLoop2
    NOP                     // (delay slot)   

    // Setup stack pointer
	MOVIA	%sp,cyg_interrupt_stack

	// Enable IRQ #0 #1 #2 to enable CWP management 
    RDCTL   %g0                 //read STATUS register
    ANDNIP  %g0, 0x7e00         //clear IPRI
    ORIP    %g0, 0x8600         //set IPRI = 3 & IE = 1
    WRCTL   %g0                 //write to STATUS register
    NOP                         //(delay slot)

    
#ifdef CYG_HAL_STARTUP_ROM
    // Copy data from ROM to RAM if the program is start from ROM rather than RAM
	.extern	hal_copy_data
   	_BSR    hal_copy_data,%g7
    NOP                    // (delay slot)
#endif

    // Zero BSS
	.extern	hal_zero_bss
   	_BSR    hal_zero_bss,%g7
    NOP                    // (delay slot)

    //platfrom initialize (initial hardware)
	.extern	hal_platform_init
   	_BSR    hal_platform_init,%g7
    NOP                    // (delay slot)
    
    // Call constructors 
	.extern	hal_invoke_constructors
   	_BSR    hal_invoke_constructors,%g7
    NOP                    // (delay slot)

    
#ifdef CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
	.extern	initialize_stub
   	_BSR    initialize_stub,%g7
    NOP                    // (delay slot)
#endif

    // Call cyg_start of the ecos
	.extern	cyg_start
   	_BSR    cyg_start,%g7
    NOP                    // (delay slot)
    
    // Loop forever if return
_LoopForever:
    BR      _LoopForever
    NOP

    
//-----------------------------------------------------------------------------
// backup_all_cwps_to_stack()
//
// backup_all_cwps_to_stack(), %sp = stack
// This function use %g3,%g4,%g5,%g7 register
// And it maybe destory status & %sp registers.

    .globl  backup_all_cwps_to_stack
backup_all_cwps_to_stack:

    // save return address to %g7
    MOV     %g7, %o7
	// retrieve the HI_LIMIT into %g5
	PFX 2
	RDCTL %g5
    ANDIP %g5,0x3e0             ; only resrve HI_LIMIT
    LSRI  %g5,1                 ; Let HI_LIMIT go to corresponding location of CWP
 
    MOV   %g3,%sp
_Int_SaveLoop:
	// reach the HI_LIMIT register window?
    RDCTL %g4
    ANDIP %g4, 0x1f0          ; only reserve CWP
    
    // compare HI_LIMIT & CWP, if equal, then exit loop
    CMP   %g5, %g4
    IFS   cc_eq
    BR    _Int_SaveDone
    NOP    

	// goto upper register window
	RESTORE
	
	// save the %lx register to stack
    ST  [%g3], %l0      // Save %l0 first
    MOV %l0  , %g3      // let %l0 = previous %sp
    MOV %g3  , %sp      // backup current %sp to %g3
	STP [%l0, 1], %l1
	STP [%l0, 2], %l2
	STP [%l0, 3], %l3
	STP [%l0, 4], %l4
	STP [%l0, 5], %l5
	STP [%l0, 6], %l6
	STP [%l0, 7], %l7

	// save the %ix registers to stack
	STP [%l0, 8], %i0
	STP [%l0, 9], %i1
	STP [%l0, 10], %i2
	STP [%l0, 11], %i3
	STP [%l0, 12], %i4
	STP [%l0, 13], %i5
	STP [%l0, 14], %i6
	STP [%l0, 15], %i7
    
	// next loop
	BR _Int_SaveLoop
	NOP

_Int_SaveDone:
    // return
    JMP %g7
    NOP

    
//-----------------------------------------------------------------------------
// restore_all_cwps_from_stack()
//
// restore_all_cwps_from_stack(), 
// %sp = input window's sp, %fp = input window's fp
// This function use %g3,%g4,%g5,%g7 register
// And it maybe destory status & %sp registers.

    .globl  restore_all_cwps_from_stack
restore_all_cwps_from_stack:

    // save return address to %g7
    MOV     %g7, %o7
    // Restore window value , if CWP less than HI_LIMIT
	PFX 2                        ; let %g5 = HI_LIMIT 
	RDCTL %g5
    ANDIP %g5,0x3e0             ; only resrve HI_LIMIT
    LSRI  %g5,1                 ; Let HI_LIMIT go to corresponding location of CWP

    MOV   %g3,%sp    
_Int_LoadLoop:
	// reach the HI_LIMIT register window?
    RDCTL %g4
    ANDIP %g4, 0x1f0          ; only reserve CWP
    
    // compare HI_LIMIT & CWP, if equal, then exit loop
    CMP   %g5, %g4
    IFS   cc_eq
    BR    _Int_LoadDone
    NOP    

	// goto upper register window
	RESTORE
	
	// load the %l1~7 register from stack
    MOV   %l0, %g3
    MOV   %g3, %sp
	LDP %l1, [%l0, 1]
	LDP %l2, [%l0, 2]
	LDP %l3, [%l0, 3]
	LDP %l4, [%l0, 4]
	LDP %l5, [%l0, 5]
	LDP %l6, [%l0, 6]
	LDP %l7, [%l0, 7]

	// load the %ix register from stack
	LDP %i0, [%l0, 8]
	LDP %i1, [%l0, 9]
	LDP %i2, [%l0, 10]
	LDP %i3, [%l0, 11]
	LDP %i4, [%l0, 12]
	LDP %i5, [%l0, 13]
	LDP %i6, [%l0, 14]
	LDP %i7, [%l0, 15]
    
    // load %l0 last
	LDP %l0, [%l0, 0]    
    
	// next loop
	BR _Int_LoadLoop
	NOP
    
_Int_LoadDone:
    // return
    JMP %g7
    NOP

    
//-----------------------------------------------------------------------------
// Interrupt Table
//
// These tables contain the isr, data and object pointers used to deliver
// interrupts to user code.
// hal_interrupt_level contains the interrupt level set by 
// HAL_INTERRUPT_CONFIGURE().
// This is a default set that provide support only for the 6 external
// interrupts in the status/cause registers. Platforms or boards are expected
// to define their own versions of these if they have their own interrupt mappings.

	.data
//	.section ".data"
	.balign 4    
    
	.extern hal_default_isr
    
	.globl	hal_interrupt_handlers
hal_interrupt_handlers:
    .rept   CYGNUM_HAL_ISR_COUNT
	.long	hal_default_isr@h
    .endr

	.globl	hal_interrupt_data
hal_interrupt_data:
	.rept	CYGNUM_HAL_ISR_COUNT
	.long	0
	.endr

	.globl	hal_interrupt_objects
hal_interrupt_objects:
	.rept	CYGNUM_HAL_ISR_COUNT
	.long	0
	.endr

//-----------------------------------------------------------------------------
// VSR Table
//
	.data
	.balign 4    
    
	.extern hal_default_exception_handler
    
	.globl	hal_vsr_table
hal_vsr_table:
    .rept   CYGNUM_HAL_EXCEPTION_COUNT
	.long	hal_default_exception_handler@h
    .endr
    

//-----------------------------------------------------------------------------
// end of vectors.S


